/*
 * Copyright (C) 2001-2002 by egnite Software GmbH. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgement:
 *
 *    This product includes software developed by egnite Software GmbH
 *    and its contributors.
 *
 * THIS SOFTWARE IS PROVIDED BY EGNITE SOFTWARE GMBH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EGNITE
 * SOFTWARE GMBH OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.ethernut.de/
 *
 */

/*
 * $Log: uartavr.c,v $
 * Revision 1.15  2002/09/15 16:40:48  harald
 * Bugfix: UART1 outputs to UART0
 *
 * Revision 1.14  2002/07/03 16:45:38  harald
 * Using GCC 3.2
 *
 * Revision 1.13  2002/06/26 17:29:08  harald
 * First pre-release with 2.4 stack
 *
 */

#include <string.h>

#include <sys/atom.h>
#include <sys/event.h>
#include <sys/timer.h>
#include <sys/ifstream.h>
#include <sys/uart.h>

#include <dev/irqreg.h>
#include <dev/uartavr.h>

/*!
 * \addtogroup xgUartAvr
 */
/*@{*/

/*!
 * \brief Get raw byte from the UART receiver register.
 *
 * \param cp Pointer to the character buffer.
 *
 * \return 1 if a byte has been available, 0 otherwise.
 */
int UartAvrGetRaw(u_char *cp)
{
    if((inp(USR) & BV(RXC)) == 0) 
        return 0;
    *cp = inp(UDR);
    return 1;
}

/*!
 * \brief Put raw byte to the UART transmit register.
 *
 * \param ch Character to transmit.
 *
 * \return 1 if the byte has been stored in the transmit
 *         register, 0 if the transmitter is busy.
 */
int UartAvrPutRaw(u_char ch)
{
    if((inp(USR) & BV(UDRE)) == 0)
        return 0;
    outp(ch, UDR);
    return 1;
}

/*
 * Handle AVR UART transmit complete.
 */
static void TxComplete(void *arg)
{
    NUTDEVICE *dev = (NUTDEVICE *)arg;
    IFSTREAM *ifs = dev->dev_icb;
    UARTDCB *dcb;

    if(ifs->if_tx_idx != ifs->if_wr_idx) {
#ifdef UDR1
        if(dev->dev_base)
            UDR1 = ifs->if_tx_buf[ifs->if_tx_idx];
        else
#endif
            outp(ifs->if_tx_buf[ifs->if_tx_idx], UDR);
        ifs->if_tx_idx++;
    }
    else {
        ifs->if_tx_act = 0;
        dcb = dev->dev_dcb;
        NutEventPostAsync(&dcb->dcb_tx_rdy);
    }
}

/*
 * Handle AVR UART receive complete.
 *
 * Note, that this function modifies the receive index in 
 * interrupt context. This requires, that any non atomic 
 * access of this index requires interrupts being disabled. 
 * Thanks to Mike Cornelius, who pointed this out.
 */
static void RxComplete(void *arg)
{
    NUTDEVICE *dev = (NUTDEVICE *)arg;
    IFSTREAM *ifs = dev->dev_icb;
    UARTDCB *dcb;

#ifdef UDR1
    if(dev->dev_base)
        ifs->if_rx_buf[ifs->if_rx_idx] = inp(UDR1);
    else
#endif
        ifs->if_rx_buf[ifs->if_rx_idx] = inp(UDR);

    if(ifs->if_rd_idx == ifs->if_rx_idx++) {
        dcb = dev->dev_dcb;
        NutEventPostAsync(&dcb->dcb_rx_rdy);
    }
}

/*!
 * \brief Wait for input.
 *
 * This function checks the input buffer for any data. If
 * the buffer is empty, the calling \ref xrThread "thread" 
 * will be blocked until at least one new character is 
 * received or a timeout occurs.
 *
 * \param dev Indicates the UART device.
 *
 * \return 0 on success, -1 on timeout.
 */
int UartAvrInput(NUTDEVICE *dev)
{
    int rc = 0;
    IFSTREAM *ifs = dev->dev_icb;
    UARTDCB *dcb;

    NutEnterCritical();
    if(ifs->if_rd_idx == ifs->if_rx_idx) {
        dcb = dev->dev_dcb;
        /*
         * Changing if into a while loop fixes a serious bug: 
         * Previous receiver events without any waiting thread 
         * set the event handle to the signaled state. So the
         * wait returns immediately. Unfortunately the calling
         * routines rely on a filled buffer when we return 0.
         * Thanks to Mike Cornelius, who found this bug.
         */
        do {
            rc = NutEventWait(&dcb->dcb_rx_rdy, dcb->dcb_rtimeout);
        } while(rc == 0 && ifs->if_rd_idx == ifs->if_rx_idx);
    }
    NutExitCritical();

    return rc;
}

/*!
 * \brief Initiate output.
 *
 * This function checks the output buffer for any data. If
 * the buffer contains at least one character, the transmitter
 * is started, if not already running. The function returns
 * immediately, without waiting for the character being
 * completely transmitted. Any remaining characters in the
 * output buffer are transmitted in the background.
 *
 * \param dev Indicates the UART device.
 *
 * \return 0 on success, -1 otherwise.
 */
int UartAvrOutput(NUTDEVICE *dev)
{
    IFSTREAM *ifs = dev->dev_icb;

    if(ifs->if_tx_act == 0 && ifs->if_tx_idx != ifs->if_wr_idx) {
        ifs->if_tx_act = 1;
#ifdef UDR1
        if(dev->dev_base)
            UDR1 = ifs->if_tx_buf[ifs->if_tx_idx];
        else
#endif
            outp(ifs->if_tx_buf[ifs->if_tx_idx], UDR);
        ifs->if_tx_idx++;
    }
    return 0;
}

/*!
 * \brief Wait for output buffer empty.
 *
 * If the output buffer contains any data, the calling
 * thread is suspended until all data has been transmitted.
 *
 * \param dev Indicates the UART device.
 *
 * \return 0 on success, -1 otherwise.
 */
int UartAvrFlush(NUTDEVICE *dev)
{
    IFSTREAM *ifs = dev->dev_icb;
    UARTDCB *dcb = dev->dev_dcb;

    /*
     * Start any pending output.
     */
    if(UartAvrOutput(dev))
        return -1;

    /*
     * Wait until output buffer empty.
     */
    NutEnterCritical();
    if(ifs->if_tx_idx != ifs->if_wr_idx)
        NutEventWait(&dcb->dcb_tx_rdy, 0);
    NutExitCritical();

    return 0;
}

/*
 * \param dev Indicates the UART device.
 *
 * \return 0 on success, -1 otherwise.
 */
static int UartAvrSetSpeed(UARTDCB *dcb, u_long baud, u_char devnum)
{
    /*
     * Refined calculation provided by Robert Hildebrand.
     */
    dcb->dcb_baudSelect = ((((2 * NutGetCpuClock()) / (baud * 16L)) + 1) / 2) - 1;

#ifdef UBRR1L
    if(devnum)
        UBRR1L = dcb->dcb_baudSelect;
    else
#endif
        outp(dcb->dcb_baudSelect, UBRR);

    return 0;
}

/*
 *
 * \param dev Indicates the UART device.
 *
 * \return 0 on success, -1 otherwise.
 */
static int UartAvrGetSpeed(UARTDCB *dcb, u_long *baud)
{
    *baud = NutGetCpuClock() / (16 * (dcb->dcb_baudSelect + 1));

    return 0;
}

/*
 *
 * \param dev Indicates the UART device.
 *
 * \return 0 on success, -1 otherwise.
 */
static int UartAvrGetStatus(NUTDEVICE *dev, u_long *status)
{
    IFSTREAM *ifs = dev->dev_icb;
    u_char us;

    *status = 0;

#ifdef UDR1
    if(dev->dev_base)
        us = inp(UCSR1A);
    else
#endif
        us = inp(USR);
    if(us & FE)
        *status |= UART_FRAMINGERROR;
    if(us & OVR)
        *status |= UART_OVERRUNERROR;
    if(ifs->if_tx_idx == ifs->if_wr_idx)
        *status |= UART_TXBUFFEREMPTY;
    if(ifs->if_rd_idx == ifs->if_rx_idx)
        *status |= UART_RXBUFFEREMPTY;
    return 0;
}

/*!
 * \brief Perform on-chip UART control functions.
 *
 * \param dev  Identifies the device that receives the device-control
 *             function. This pointer must have been retrieved previously
 *             by using the NutDeviceOpen() function.
 * \param req  Requested control function. May be set to one of the
 *             following constants:
 *             - UART_SETSPEED, if conf points to an u_long value containing the baudrate.
 *             - UART_GETSPEED, if conf points to an u_long value receiving the current baudrate.
 *             - UART_SETDATABITS
 *             - UART_GETDATABITS
 *             - UART_SETPARITY
 *             - UART_GETPARITY
 *             - UART_SETSTOPBITS
 *             - UART_GETSTOPBITS
 *             - UART_SETSTATUS
 *             - UART_GETSTATUS
 *             - UART_SETREADTIMEOUT
 *             - UART_GETREADTIMEOUT
 *             - UART_SETWRITETIMEOUT
 *             - UART_GETWRITETIMEOUT
 *
 * \param conf Points to a buffer that contains any data required for
 *             the given control function or receives data from that
 *             function.
 * \return 0 on success, -1 otherwise.
 *
 * \note Timeout is limited to the granularity of the system timer.
 *
 */
int UartAvrIOCtl(NUTDEVICE *dev, int req, void *conf)
{
    int rc = -1;
    UARTDCB *dcb;
        
    if(dev == 0)
        dev = &devUart0;
    
    dcb = dev->dev_dcb;

    switch(req) {
    case UART_SETSPEED:
        rc = UartAvrSetSpeed(dcb, *((u_long *)conf), dev->dev_base);
        break;
    case UART_GETSPEED:
        rc = UartAvrGetSpeed(dcb, (u_long *)conf);
        break;
    case UART_SETDATABITS:
        if(*((u_long *)conf) == 8)
            rc = 0;
        break;
    case UART_GETDATABITS:
        *((u_long *)conf) = 8;
        rc = 0;
        break;
    case UART_SETPARITY:
        if(*((u_long *)conf) == 0)
            rc = 0;
        break;
    case UART_GETPARITY:
        *((u_long *)conf) = 0;
        rc = 0;
        break;
    case UART_SETSTOPBITS:
        if(*((u_long *)conf) == 1)
            rc = 0;
        break;
    case UART_GETSTOPBITS:
        *((u_long *)conf) = 1;
        rc = 0;
        break;
    case UART_GETSTATUS:
        UartAvrGetStatus(dev, (u_long *)conf);
        break;
    case UART_SETSTATUS:
        break;
    case UART_SETREADTIMEOUT:
        dcb->dcb_rtimeout = *((u_long *)conf);
        rc = 0;
        break;
    case UART_GETREADTIMEOUT:
        *((u_long *)conf) = dcb->dcb_rtimeout;
        rc = 0;
        break;
    case UART_SETWRITETIMEOUT:
        dcb->dcb_wtimeout = *((u_long *)conf);
        rc = 0;
        break;
    case UART_GETWRITETIMEOUT:
        *((u_long *)conf) = dcb->dcb_wtimeout;
        rc = 0;
        break;
    }
    return rc;
}

/*!
 * \brief Initialize on chip uart device.
 *
 * Prepares the device for subsequent reading or writing.
 * Enables UART transmitter and receiver interrupts.
 *
 * Applications should not use this function, but call
 * NutDeviceOpen().
 *
 * \param dev  Identifies the device to initialize.
 *
 * \return 0 on success, -1 otherwise.
 */
int UartAvrInit(NUTDEVICE *dev)
{
    IFSTREAM *ifs;
    UARTDCB *dcb;

    /*
     * We only support character devices for on-chip UARTs.
     */
    if(dev->dev_type != IFTYP_STREAM || dev->dev_irq != 0 || 
       (dev->dev_base != 0 && dev->dev_base != 1))
        return -1;

    /*
     * Initialize interface control block.
     */
    ifs = dev->dev_icb;
    memset(ifs, 0, sizeof(IFSTREAM));
    ifs->if_input = UartAvrInput;
    ifs->if_output = UartAvrOutput;
    ifs->if_flush = UartAvrFlush;

    /*
     * Initialize driver control block.
     */
    dcb = dev->dev_dcb;
    memset(dcb, 0, sizeof(UARTDCB));
    dcb->dcb_baudSelect = 1;

    /*
     * Register interrupt handler.
     */
    if(dev->dev_base) {
#ifdef UDR1
        if(NutRegisterInterrupt(IRQ_UART1_RX, RxComplete, dev))
            return -1;
        if(NutRegisterInterrupt(IRQ_UART1_TX, TxComplete, dev))
            return -1;
#endif
    }
    else {
        if(NutRegisterInterrupt(IRQ_UART_RX, RxComplete, dev))
            return -1;
        if(NutRegisterInterrupt(IRQ_UART_TX, TxComplete, dev))
            return -1;
    }

    /*
     * Set baudrate and handshake default.
     */
    UartAvrSetSpeed(dcb, 9600, dev->dev_base);

    /*
     * Enable interrupts.
     */
#ifdef UCSR1B
    if(dev->dev_base)
        UCSR1B = BV(RXCIE) | BV(TXCIE) | BV(RXEN) | BV(TXEN);
    else
#endif
        outp(BV(RXCIE) | BV(TXCIE) | BV(RXEN) | BV(TXEN), UCR);
        

    return 0;
}


/*@}*/
