/*
 * Copyright (C) 2001-2002 by egnite Software GmbH. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgement:
 *
 *    This product includes software developed by egnite Software GmbH
 *    and its contributors.
 *
 * THIS SOFTWARE IS PROVIDED BY EGNITE SOFTWARE GMBH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EGNITE
 * SOFTWARE GMBH OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.ethernut.de/
 *
 */

/*
 * $Log: device.c,v $
 * Revision 1.17  2002/09/15 17:07:54  harald
 * Do not override base and irq set to 0
 *
 * Revision 1.16  2002/09/03 17:39:47  harald
 * Comment about broken connection
 *
 * Revision 1.15  2002/06/26 17:29:44  harald
 * First pre-release with 2.4 stack
 *
 */

#include <string.h>

#include <sys/ifstream.h>
#include <sys/uart.h>

#ifdef NUTDEBUG
#include <sys/print.h>
#endif

/*!
 * \addtogroup xgDevice
 */
/*@{*/

/*!
 * \brief Linked list of all registered devices.
 */
NUTDEVICE *nutDeviceList = 0;

/*!
 * \brief Find device entry by name.
 *
 * \param  name Unique device name.
 *
 * \return Pointer to the ::NUTDEVICE structure.
 */
NUTDEVICE *NutDeviceLookup(CONST char *name)
{
    NUTDEVICE *dev;

    for(dev = nutDeviceList; dev; dev = dev->dev_next)
        if(strcmp(dev->dev_name, name) == 0)
            break;

    return dev;
}

/*!
 * \brief Register a device.
 *
 * Initializes the device and adds it to the system device list.
 * Applications should call this function during initialization
 * for each device they intend to use. Once registered, stream
 * devices must be accessed by calling NutDeviceOpen() first.
 * However, network devices should not be opened but configured
 * by calling NutNetIfConfig() or NutNetAutoConfig().
 *
 * If called for a device, which had been registered previously,
 * this function will update the base address and interrupt
 * number. However, most device drivers do not expect these
 * itemes being changed and may crash.
 *
 * \note Virtual stream devices don't need to be registered.
 *
 * \param dev  Pointer to the ::NUTDEVICE structure, which is
 *             provided by the device driver. Should point to
 *              - devUart0 for the first on-chip UART.
 *              - devUart1 for the second on-chip UART.
 *              - devUarts[0-7] for one of the 8 SPI UARTs.
 *              - devEth0 for the Ethernet device.
 * \param base Hardware base address of this device. Set to 0,
 *             if the device driver has a fixed hardware address.
 * \param irq  Hardware interrupt used by this device. Set to 0,
 *             if the device driver doesn't support configurable
 *             interupts.
 * \return 0 if the device has been registered for the first time.
 *         The function returns -1 if the device had been registered
 *         previously or the ::NUTDEVICE structure is invalid.
 */
int NutRegisterDevice(NUTDEVICE *dev, u_short base, u_char irq)
{
    int rc = -1;

    if((void *)dev >= RAMSTART) {
        if(base)
            dev->dev_base = base;
        if(irq)
            dev->dev_irq = irq;

        if(NutDeviceLookup(dev->dev_name) == 0) {
            dev->dev_next = nutDeviceList;
            nutDeviceList = dev;
            rc = 0;
        }
    }
    return rc;
}

/*!
 * \brief Open a stream device by name.
 *
 * This function must be called after registering the device with
 * NutRegisterDevice() and before calling any read or write
 * function.
 *
 * \note Do not use this function for network devices. For these 
 *       you must call NutNetIfConfig() or NutNetAutoConfig().
 *       Also virtual stream devices need not to be opened.
 *
 * \param name Case sensitiv name of the device. This name is fixed by the 
 *             device driver and may be any of the following:
 *              - "uart0" for the first on-chip UART.
 *              - "uart1" for the second on-chip UART.
 *              - "uarts0".."uarts7" for one of the 8 SPI UARTs.
 * \return Pointer to the device structure or null if the device
 *         has not been registered.
 */
NUTDEVICE *NutDeviceOpen(CONST char *name)
{
    NUTDEVICE *dev;

    if((dev = NutDeviceLookup(name)) != 0) {
        if(dev->dev_type == IFTYP_STREAM) {
            if((*dev->dev_init)(dev)) 
                return 0;
            NutDeviceRead(dev, 0, 0);
        }
    }
    return dev;
}

/*!
 * \brief Closes a previously opened stream device.
 *
 * The calling thread may be suspended until all buffered 
 * output data has been sent.
 *
 * \param dev Identifies the device to close. This pointer
 *            must have been retrieved by calling NutDeviceOpen().
 *
 * \return 0 on success, -1 otherwise.
 */
int NutDeviceClose(NUTDEVICE *dev)
{
    /*
     * Flush our output buffer.
     */
    if(dev->dev_type == IFTYP_STREAM)
        return NutIfStreamFlush(dev);

    return -1;
}

/*!
 * \brief Read up to a specified number of bytes from a device.
 *
 * The function may read fewer than the given number of bytes.
 *
 * \param dev  Identifies the device to read from. A null pointer may be used 
 *             for unbuffered input from the first on-chip UART. If this 
 *             pointer is not null, it must have been retrieved by directly 
 *             or indirectly calling NutDeviceOpen() for real devices. For 
 *             virtual devices this pointer is returned by the function that 
 *             creates the device.
 * \param data Pointer to the buffer that receives the data.
 * \param size Size of the buffer.
 *
 * \return The number of bytes read, 0 on timeouts and broken
 *         connections or -1 in case of an error.
 */
int NutDeviceRead(NUTDEVICE *dev, void *data, int size)
{
    if(dev == 0) {
        if(data >= RAMSTART && size > 0)
            return UartAvrGetRaw(data);
    }
    else if(dev->dev_type == IFTYP_STREAM)
        return NutIfStreamRead(dev, data, size);

    return -1;
}

/*!
 * \brief Read up to a specified number of bytes from a device 
 *        translating non printables.
 *
 * This call works like NutDeviceRead(), but translates
 * character codes below 33, above 126 or equal 34 into
 * their hex representation prepended by a backslash and the
 * letter x. Backslashes are duplicated and the complete string
 * is terminated by character code zero. This offers the ability 
 * to store binary data in printable strings.
 *
 * \todo Very specific code, should be moved to an extra module
 *       to save ROM space of ordinary applications.
 *
 * \param dev Identifies the device that read from.
 *            A null pointer specifies the first on-chip UART.
 *            If this pointer is not null, it must have been 
 *            retrieved by using the NutDeviceOpen() function.
 * \param data Pointer to the buffer that receives the translated string.
 * \param size Size of the buffer.
 *
 * \return The number of bytes read, 0 on timeouts and broken
 *         connections or -1 in case of an error. Due to code 
 *         translation, the number of bytes read be less than 
 *         the number of characters returned in the string.
 */
int NutDeviceReadTran(NUTDEVICE *dev, void *data, int size)
{
    if(dev->dev_type == IFTYP_STREAM)
        return NutIfStreamReadTran(dev, data, size);

    return -1;
}

/*!
 * \brief Get a line from a specified device.
 *
 * Reads a string from the specified device. Characters are
 * read up to and including the first newline character, up
 * to the end of the stream, or until the number of characters
 * read is equal to the specified size, whichever comes first.
 *
 * \param dev  Identifies the device to read from. A null pointer may be used 
 *             for unbuffered output to the first on-chip UART. If this 
 *             pointer is not null, it must have been retrieved by directly 
 *             or indirectly calling NutDeviceOpen() for real devices. For 
 *             virtual devices this pointer is returned by the function that 
 *             creates the device.
 * \param data Pointer to the buffer that receives the data.
 * \param size Size of the buffer.
 *
 * \return The number of bytes read or -1 in case of an error.
 */
int NutDeviceGetLine(NUTDEVICE *dev, void *data, int size)
{
    if(dev->dev_type == IFTYP_STREAM)
        return NutIfStreamGetLine(dev, data, size);

    return -1;
}

/*!
 * \brief Write a buffer to a specified device.
 *
 * This is a raw output without any character translation
 * like EOL (end of line).
 *
 * \param dev  Identifies the device to write to. A null pointer may be used 
 *             for unbuffered output to the first on-chip UART. If this 
 *             pointer is not null, it must have been retrieved by directly 
 *             or indirectly calling NutDeviceOpen() for real devices. For 
 *             virtual devices this pointer is returned by the function that 
 *             creates the device.
 * \param data Buffer to be written.
 * \param len  Number of characters to be printed.
 *
 * \return If successful, the number of bytes added to the
 *         output buffer. This may be less than the specified 
 *         number of bytes to print. The return value -1 
 *         indicates an error.
 */
int NutDeviceWrite(NUTDEVICE *dev, CONST void *data, int len)
{
    if(dev == 0) {
        if(data >= RAMSTART && len > 0)
            return UartAvrPutRaw(*((char *)data));
    }
    else if(dev->dev_type == IFTYP_STREAM)
        return NutIfStreamWrite(dev, data, len);

    return -1;
}

int NutDeviceWriteTran(NUTDEVICE *dev, CONST void *data)
{
    if(dev->dev_type == IFTYP_STREAM)
        return NutIfStreamWriteTran(dev, data);

    return -1;
}

/*!
 * \brief Write a buffer in program space to a specified device.
 *
 * This is a raw output without any character translation
 * like EOL (end of line).
 *
 * \param dev  Identifies the device to write to. A null pointer may be used 
 *             for unbuffered output to the first on-chip UART. If this 
 *             pointer is not null, it must have been retrieved by directly 
 *             or indirectly calling NutDeviceOpen() for real devices. For 
 *             virtual devices this pointer is returned by the function that 
 *             creates the device.
 * \param data Buffer to be written.
 * \param len  Number of characters to be printed.
 *
 * \return If successful, the number of bytes added to the
 *         output buffer. This may be less than the specified 
 *         number of bytes to print. The return value -1 
 *         indicates an error.
 */
int NutDeviceWrite_P(NUTDEVICE *dev, PGM_P data, int len)
{
    if(dev == 0) {
        if(len > 0)
            return UartAvrPutRaw(PRG_RDB(data));
    }
    else if(dev->dev_type == IFTYP_STREAM)
        return NutIfStreamWrite_P(dev, data, len);

    return -1;
}

/*!
 * \brief Perform device specific control functions.
 *
 * \param dev  Identifies the device to control. This pointer must have 
 *             been retrieved by directly or indirectly calling 
 *             NutDeviceOpen().
 * \param req  Requested control function.
 * \param conf Points to a buffer that contains any data required for
 *             the given control function or receives data from that
 *             function.
 */
int NutDeviceIOCtl(NUTDEVICE *dev, int req, void *conf)
{
    if(dev == 0)
        return UartAvrIOCtl(0, req, conf);
    else if((void *)dev >= RAMSTART && dev->dev_ioctl)
        return (*dev->dev_ioctl)(dev, req, conf);

    return -1;
}
/*@}*/
