/*
 * Copyright (C) 2001-2002 by egnite Software GmbH. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgement:
 *
 *    This product includes software developed by egnite Software GmbH
 *    and its contributors.
 *
 * THIS SOFTWARE IS PROVIDED BY EGNITE SOFTWARE GMBH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EGNITE
 * SOFTWARE GMBH OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.ethernut.de/
 *
 * -
 * Portions Copyright (C) 2000 David J. Hudson <dave@humbug.demon.co.uk>
 *
 * This file is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.
 *
 * You can redistribute this file and/or modify it under the terms of the GNU
 * General Public License (GPL) as published by the Free Software Foundation;
 * either version 2 of the License, or (at your discretion) any later version.
 * See the accompanying file "copying-gpl.txt" for more details.
 *
 * As a special exception to the GPL, permission is granted for additional
 * uses of the text contained in this file.  See the accompanying file
 * "copying-liquorice.txt" for details.
 */

/*
 * $Log: event.c,v $
 * Revision 1.16  2002/06/26 17:29:44  harald
 * First pre-release with 2.4 stack
 *
 */

#include <sys/atom.h>
#include <sys/timer.h>
#include <sys/thread.h>
#include <sys/event.h>

#ifdef NUTDEBUG
#include <sys/print.h>
#include <sys/osdebug.h>
#endif

/*!
 * \addtogroup xgEvent
 */
/*@{*/


/*!
 * \brief Timer callback in case of event timeout.
 *
 * \param timer Handle of the elapsed timeout timer.
 * \param arg   Handle of an event queue.
 *
 * \note This routine is running in interrupt context.
 */
void NutEventTimeout(HANDLE timer, void *arg)
{
    NUTTHREADINFO *tqp;
    NUTTHREADINFO * volatile *tqpp = arg;

    /*
     * A signaled queue is an empty queue. So our
     * thread already left this queue.
     */
    if((tqp = *tqpp) != SIGNALED) {

        /*
         * Walk down the linked list and identify
         * the thread by the timer id it is waiting
         * for.
         */
        while(tqp) {
            if(tqp->td_timer == timer) {
                /*
                 * Found the thread. Remove it from the
                 * event queue and add it to the queue
                 * of threads, which are ready to run.
                 */
                *tqpp = tqp->td_qnxt;
                tqp->td_qnxt = 0;
                tqp->td_timer = 0;
                NutThreadResumeAsync(tqp);
                break;
            }
            tqpp = &tqp->td_qnxt;
            tqp = tqp->td_qnxt;
        }
    }
}

/*!
 * \brief Wait for an event in a specified queue.
 *
 * Give up the CPU until another thread posts an event to this
 * queue or until a time-out occurs, whichever comes first.
 * If previously an event had been posted to this queue without 
 * any thread waiting, then the thread will continue running,
 * if no other thread with higher priority is ready to run.
 *
 * \param qhp Identifies the queue to wait on.
 * \param ms  Maximum wait time in milliseconds.
 *
 * \return 0 if event received, -1 on timeout.
 *
 * \note Timeout is limited to the granularity of the system timer.
 */
int NutEventWait(volatile HANDLE *qhp, u_long ms)
{
    NutEnterCritical();

    /*
     * Check for posts on a previously empty queue. 
     */
    if(*qhp == SIGNALED) {
        //if(runningThread == app1 || runningThread == app2 || runningThread == app3 || runningThread == app4)
        //    NutPrintFormat(0, "(%04X-%04X)", runningThread, qhp);

        /*
         * Even if already signaled, switch to any other thread, which 
         * is ready to run and has the same or higher priority.
         */
        NutExitCritical();
        NutThreadYield();
        NutEnterCritical();
        if(*qhp == SIGNALED)
            *qhp = 0;
        NutExitCritical();
        return 0;
    }

    /*
     * Remove the current thread from the list
     * of running threads and add it to the
     * specified queue.
     */
#ifdef NUTDEBUG
    if(os_trace)
        NutPrintFormat(0, "Rem<%04x>", runningThread);
#endif
    NutThreadRemoveQueue(runningThread, &runQueue);
#ifdef NUTDEBUG
    if(os_trace)
        NutKDumpThreadList();
        //NutKDumpThreadQueue(runQueue);
#endif
    runningThread->td_state = TDS_SLEEP;
    //if(runningThread == app1 || runningThread == app2 || runningThread == app3 || runningThread == app4)
    //    NutPrintFormat(0, "[%04X-%04X]", runningThread, qhp);
    NutThreadAddPriQueue(runningThread, (NUTTHREADINFO **)qhp);

    /*
     * If a timeout value had been given, start
     * a oneshot timer.
     */
    if(ms)
        runningThread->td_timer = NutTimerStart(ms, NutEventTimeout, (void *)qhp, TM_ONESHOT);
    else
        runningThread->td_timer = 0;

    /*
     * Switch to the next thread, which is ready
     * to run.
     */
#ifdef NUTDEBUG
    if(os_trace)
        NutPrintFormat(0, "SWW<%04x %04x>", runningThread, runQueue);
#endif
    NutThreadSwitch();

    /*
     * If our timer handle is still set, we were
     * woken up by an event. If the handle is
     * cleared and a timeout value had been
     * specified, then we know that a timeout
     * occured.
     */
    if(runningThread->td_timer)
        runningThread->td_timer = 0;
    else if(ms) {
        NutExitCritical();
        return -1;
    }
    NutExitCritical();
    return 0;
}

/*!
 * \brief Wait for a new event in a specified queue.
 *
 * Give up the CPU until another thread posts an event to this
 * queue or until a time-out occurs, whichever comes first.
 * Previously posted events to an empty queue will be ignored.
 *
 * \param qhp Identifies the queue to wait on.
 * \param ms  Maximum wait time in milliseconds.
 *
 * \return 0 if event received, -1 on timeout.
 *
 * \note Timeout is limited to the granularity of the system timer.
 */
int NutEventWaitNext(volatile HANDLE *qhp, u_long ms)
{
    int rc;

    NutEnterCritical();

    /*
     * Check for posts on a previously empty queue. 
     */
    if(*qhp == SIGNALED)
        *qhp = 0;

    rc = NutEventWait(qhp, ms);

    NutExitCritical();

    return rc;
}

/*!
 * \brief Asynchronously post an event to a specified queue.
 *
 * Wake up the thread with the highest priority waiting on
 * the specified queue. But even if the priority of the woken
 * thread is higher than the current thread's priority, the
 * current one continues running.
 *
 * \note It is save to call this function from within an interrupt
 *       handler. In any case interrupts should be disabled when
 *       calling this function.
 *
 * \param qhp Identifies the queue an event is posted to.
 *
 */
void NutEventPostAsync(HANDLE volatile *qhp)
{
    NUTTHREADINFO *td;

    if(*qhp != SIGNALED) {
        if((td = *qhp) != 0) {
            NutThreadRemoveQueue(td, (NUTTHREADINFO *volatile *)qhp);

            /*
             * Stop any running timeout timer.
             */
            if(td->td_timer)
                NutTimerStopAsync(td->td_timer);
            NutThreadResumeAsync(td);
        }
        else 
            *qhp = SIGNALED;
    }
}

/*!
 * \brief Post an event to a specified queue.
 *
 * Wake up the thread with the highest priority waiting on
 * this queue. If the priority of the waiting thread is higher
 * or equal than the current thread's priority, then the
 * current thread is stopped and CPU control is passed
 * to the waiting thread.
 *
 * \param qhp Identifies the queue an event is posted to.
 *
 */
void NutEventPost(HANDLE *qhp)
{
    NutEnterCritical();

    NutEventPostAsync(qhp);

    /*
     * If any thread with higher priority is
     * ready to run, switch the context.
     */
    if(runningThread != runQueue) {
        runningThread->td_state = TDS_READY;
#ifdef NUTDEBUG
        if(os_trace)
            NutPrintFormat(0, "SWP<%04x %04x>", runningThread, runQueue);
#endif
        NutThreadSwitch();
    }
    NutExitCritical();
}

/*!
 * \brief Asynchronously broadcast an event to a specified queue.
 *
 * Wake up all threads waiting on this queue. But even if the
 * priority of any woken thread is higher than the current thread's
 * priority, the current one continues running.
 *
 * \note It is save to call this function from within an interrupt
 *       handler. In any case interrupts must be disabled when
 *       calling this function.
 *
 * \param qhp Identifies the queue an event is broadcasted to.
 *
 */
void NutEventBroadcastAsync(HANDLE *qhp)
{
    if(*qhp != SIGNALED)
        while(*qhp)
            NutEventPostAsync(qhp);
}

/*!
 * \brief Broadcast an event to a specified queue.
 *
 * Wake up all threads waiting on this queue. If the
 * priority of any waiting thread is higher or equal
 * than the current thread's priority, then the
 * current thread is stopped and CPU control is passed
 * to the woken up thread with the highest priority.
 *
 * \param qhp Identifies the queue an event is broadcasted to.
 *
 */
void NutEventBroadcast(HANDLE *qhp)
{
    if(*qhp != SIGNALED)
        while(*qhp)
            NutEventPost(qhp);
}

/*@}*/
